mod dirs;
mod tui;


use std::fs;
use std::io::Write;
use std::path::PathBuf;

use anyhow::{ensure, Result};
use clap::{Parser, Subcommand};


#[derive(Debug, Parser)]
pub struct Args {
    #[command(subcommand)]
    cmd: Option<Command>,
}

#[derive(Debug, Subcommand)]
enum Command {
    /// Add a project
    Add { path: PathBuf },

    /// List existing projects
    List,
}


#[derive(Debug, Default, serde::Serialize, serde::Deserialize)]
struct Projects {
    list: Vec<PathBuf>,
}


fn main() -> Result<()> {
    let args = Args::parse();

    let projects = read_projects_file()?;

    let result = match &args.cmd {
        Some(cmd) => handle_subcommands(projects, cmd),
        None => run_tui(projects),
    };

    if let Err(err) = result {
        eprintln!("{err}");
        std::process::exit(1);
    }

    Ok(())
}


fn run_tui(projects: Projects) -> Result<()> {
    clear_selected_project_file()?;
    let selected_project = tui::run(projects)?;

    if let Some(selected_project) = selected_project {
        write_selected_project_file(selected_project)?
    }

    Ok(())
}


fn handle_subcommands(mut projects: Projects, cmd: &Command) -> Result<()> {
    match cmd {
        Command::Add { path } => {
            add_project(&mut projects, path)?;
        }
        Command::List => {
            list_projects(&projects)?;
        }
    }

    Ok(())
}


fn add_project<P>(projects: &mut Projects, path: P) -> Result<()>
where
    P: Into<PathBuf>,
{
    let path = path.into();

    ensure!(path.is_dir(), "Project path does not exists");
    ensure!(
        !projects.list.contains(&path.to_path_buf()),
        "Project path already registered"
    );

    projects.list.push(path);
    write_projects_file(projects)?;

    Ok(())
}


fn list_projects(projects: &Projects) -> Result<()> {
    for project in &projects.list {
        println!("{project:?}")
    }

    Ok(())
}

fn clear_selected_project_file() -> Result<()> {
    let cache_file = dirs::cache_path().join("selected-project");

    if fs::exists(&cache_file)? {
        fs::remove_file(cache_file)?;
    }

    Ok(())
}

fn write_selected_project_file(path: PathBuf) -> Result<()> {
    let cache_file = dirs::cache_path().join("selected-project");
    let mut file = fs::OpenOptions::new()
        .write(true)
        .create(true)
        .truncate(true)
        .open(cache_file)?;
    write!(file, "{}", path.display())?;
    Ok(())
}

fn write_projects_file(projects: &Projects) -> Result<()> {
    let projects_file = dirs::data_path().join("projects");

    let file = fs::OpenOptions::new()
        .write(true)
        .create(true)
        .truncate(true)
        .open(projects_file)?;

    bincode::serialize_into(file, projects)?;

    Ok(())
}


fn read_projects_file() -> Result<Projects> {
    let projects_file = dirs::data_path().join("projects");

    if !projects_file.exists() {
        return Ok(Projects::default());
    }

    let file = std::fs::File::open(projects_file)?;
    let projects = bincode::deserialize_from(file)?;

    Ok(projects)
}
